// This file is part of Lx.
//
// Copyright 2003, 2004, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.helpers;

import org.xml.sax.*;

import java.io.File;
import java.io.Writer;
import java.io.IOException;
import java.util.regex.Pattern;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;

/**
 * Transforms a SAX stream into an Lx input file.  This class
 * implements the <code>ContentHandler</code> class, and its action
 * on handling the SAX stream is to write a formatted Lx file to the
 * given output.
 *
 * <p>The class also has a <code>main</code> method, which consists
 * of an application which reads an input XML file and writes out a
 * corresponding Lx file.
 *
 * <p>Usage is
 * <pre>
 * java &lt;classpath&gt; uk.me.nxg.lx.helpers.LxWriter &lt;options&gt; input.xml
 * </pre>
 * Options are as follows
 * <dl>
 * <dt>--output outputfile</dt>
 * <dd>Names the output file.  If not given, then the output is
 * sent to stdout.</dd>
 * <dt>--validating</dt>
 * <dd>If present, the created XML parser is validating; if not, not.</dd>
 * <dt>--trim, --notrim</dt>
 * <dd>If present, then whitespace is trimmed from the beginning and
 * end of SAX <code>characters</code> events, and if this leaves
 * nothing, nothing is output (default <code>--trim</code>).</dd>
 * </dl>
 */
public class LxWriter
        implements org.xml.sax.ContentHandler, org.xml.sax.ErrorHandler {

    private Locator locator;
    private Writer writer;
    private int lastEvent;

    private boolean trimWhitespace = true;
    private Pattern whitespace = Pattern.compile("\\s+");
    private Pattern newlineCharacters = Pattern.compile("[\\n\\r]");
    // Last character below is a doubly-escaped backslash.
    // Characters [{}|] are in this set, not because LxReader needs
    // them to be escaped, but because emacs' lisp mode gets confused
    // if they appear outside a string.
    private Pattern quotedCharacters = Pattern.compile("[#:(){}\\[\\]\"\\|\\s\\\\]");
    private Pattern escapedCharacters = Pattern.compile("[\\[\\]\\\\]");

    private PrettyPrinter pp;

    private static final int START = 0;
    private static final int PREFIX_START = 1;
    private static final int PREFIX_END = 2;
    private static final int ELEMENT_START = 3;
    private static final int ELEMENT_END = 4;
    private static final int CHARACTERS = 5;

    private static final int GENERAL_MODE = 0;
    private static final int XSLT_MODE = 1;
    private static final String xslNamespace
            = "http://www.w3.org/1999/XSL/Transform";

    // Must be static, since invoked from main.  The `outputMode'
    // controls a couple of layout tweaks.  Nothing terribly important.
    private static int outputMode = GENERAL_MODE;

    /**
     * Creates a writer sending output to the given file.
     *
     * @param file a File to receive the generated Lx file
     * @throws IOException if the file cannot be found
     */
    public LxWriter(File file)
            throws IOException {
        writer = new java.io.FileWriter(file);
        pp = new PrettyPrinter();
    }

    /**
     * Creates an LxWriter sending output to the given Writer
     * @param writer a Writer to receive the generated Lx file
     */
    public LxWriter(Writer writer) {
        this.writer = writer;
        pp = new PrettyPrinter();
    }

    /**
     * Creates an LxWriter sending output to the given OutputStream
     * @param stream an OutputStream to receive the generated Lx file
     */
    public LxWriter(java.io.OutputStream stream) {
        writer = new java.io.PrintWriter(stream);
        assert writer != null;
        pp = new PrettyPrinter();
    }

    /**
     * Trim whitespace or not.  If this is called with a
     * <code>true</code> value, then whitespace is trimmed from the
     * beginning and end of SAX <code>characters</code> events, and if
     * this leaves nothing, nothing is output.  If this is set
     * <code>false</code>, then all whitespace is preserved.
     *
     * <p>The initial value is <code>true</code>.  Note that this is
     * generally what you want when you are converting a markup-heavy
     * XML file to <code>.lx</code> format, but it does mean that the
     * round-trip XML to Lx to XML is not an identity operation.
     *
     * @param value true if whitespace is to be trimmed from the input
     */
    public void setTrimWhitespace(boolean value) {
        trimWhitespace = value;
    }

    /* Start of ContentHandler interface */

    public void setDocumentLocator (Locator locator) {
        this.locator = locator;
    }
    
    public void startDocument ()
            throws SAXException {
        assert writer != null;
        assert pp != null;

        lastEvent = START;
    }
    
    public void endDocument()
            throws SAXException {
        try {
            writer.close();
        } catch (IOException e) {
            throw new SAXException(e);
        }
    }

    public void startPrefixMapping (String prefix, String uri)
            throws SAXException {
        try {
            StringBuffer sb = new StringBuffer();
            if (lastEvent != START)
                pp.newline(sb);

            if (lastEvent != PREFIX_START) {
                sb.append("(lx:namespace (");
                pp.push(15);    // length of above string
            }

            if (uri.equals(xslNamespace)
                && outputMode != XSLT_MODE) {
                System.err.println
                        ("XSLT namespace found: shouldn't you give option --mode=xslt?");
            }

            sb.append('(');
            if (prefix.length() == 0)
                sb.append("#f");
            else
                sb.append('"').append(prefix).append('"');
            sb.append(" \"")
                    .append(uri)
                    .append("\")");
            writer.write(sb.toString());
            lastEvent = PREFIX_START;
        } catch (IOException e) {
            throw new SAXException(e);
        }
    }

    public void endPrefixMapping (String prefix)
            throws SAXException {
        try {
            if (lastEvent != PREFIX_END) {
                writer.write(')');
                pp.pop();
                lastEvent = PREFIX_END;
            }
        } catch (IOException e) {
            throw new SAXException(e);
        }
    }

    public void startElement (String uri, String localName,
			      String qName, Attributes atts)
            throws SAXException {
        try {
            StringBuffer sb = new StringBuffer();

            switch (lastEvent) {
              case PREFIX_START:
                sb.append(')');
                pp.pop();
                pp.push(2);
                pp.newline(sb);
                break;

              case PREFIX_END:
              case ELEMENT_START:
              case ELEMENT_END:
              case CHARACTERS:
                pp.newline(sb);
                break;

            }

            if (outputMode == XSLT_MODE
                && uri.equals(xslNamespace)
                && localName.equals("template"))
                pp.newline(sb);

            sb.append('(').append(qName);
            if (atts.getLength() > 0) {
                sb.append(' ');
                pp.push(qName.length() + 2);
                for (int i=0; i<atts.getLength(); i++) {
                    if (atts.getQName(i).startsWith("xml")) 
                        // shouldn't be here, I don't think, but don't worry
                        continue;
                    if (i > 0)
                        pp.newline(sb);
                    sb.append(atts.getQName(i)).append(": ");
                    String val = atts.getValue(i);
                    if (quotedCharacters.matcher(val).find())
                        writeQuotedString(val, sb);
                    else
                        sb.append(val);
                }
                pp.pop();
            }
            pp.push(2);
            writer.write(sb.toString());
            lastEvent = ELEMENT_START;
        } catch (IOException e) {
            throw new SAXException(e);
        }
    }

    public void endElement (String uri, String localName,
			    String qName)
            throws SAXException {
        try {
            writer.write(')');
            pp.pop();
            lastEvent = ELEMENT_END;
        } catch (IOException e) {
            throw new SAXException(e);
        }
    }

    public void characters(char ch[], int start, int length)
            throws SAXException {
        try {
            String outstring;
            int end = start+length-1; // index of last character

            if (trimWhitespace) {
                for (; start <= end; start++)
                    if (!Character.isWhitespace(ch[start]))
                        break;
                for (; start <= end; end--)
                    if (!Character.isWhitespace(ch[end]))
                        break;
            }

            if (start <= end) {
                if (lastEvent == ELEMENT_START)
                    pp.newline(writer);
                else
                    writer.write(' ');

                CharSequence chs
                        = java.nio.CharBuffer.wrap(ch, start, end-start+1);
                if (quotedCharacters.matcher(chs).find()) {
                    writer.write(writeQuotedString(chs, new StringBuffer())
                                 .toString());
                } else {
                    writer.write(ch, start, end-start+1);
                }

                lastEvent = CHARACTERS;
            }

        } catch (IOException e) {
            throw new SAXException(e);
        }
    }

    public void ignorableWhitespace (char ch[], int start, int length)
            throws SAXException {
        // do nothing
    }

    public void processingInstruction (String target, String data)
            throws SAXException {
        // do nothing
    }

    public void skippedEntity (String name)
            throws SAXException {
        // do nothing
    }

    /* End of ContentHandler interface */

    /* Start of ErrorHandler interface */
    public void warning (SAXParseException exception)
            throws SAXException {
        System.err.println("SAX warning: " + exception);
        //throw new SAXException("Parse warning", exception);
    }
    
    public void error (SAXParseException exception)
            throws SAXException {
        System.err.println("SAX error: " + exception);
        //throw new SAXException("Parse error", exception);
    }
    
    public void fatalError (SAXParseException exception)
            throws SAXException {
        System.err.println("SAX fatal error: " + exception);
        //throw new SAXException("Fatal parse error", exception);
    }
    /* End of ErrorHandler interface */

    private StringBuffer writeQuotedString(CharSequence seq,
                                           StringBuffer sb) {
        sb.append('\"');
        for (int i=0; i<seq.length(); i++) {
            char ch = seq.charAt(i);
            switch (ch) {
              case '\"':
                sb.append('\\');
                break;
              case '\\':
                sb.append('\\');
                break;
              default:
                break;
            }
            sb.append(ch);
        }
        return sb.append('\"');
    }

    private static class PrettyPrinter {
        private java.util.Stack stack;

        PrettyPrinter() {
            stack = new java.util.Stack();
        }
        void push(int extraIndent) {
            if (stack.empty())
                stack.push(new Integer(extraIndent));
            else {
                Integer currTop = (Integer)stack.peek();
                stack.push(new Integer(currTop.intValue() + extraIndent));
            }
        }
        void pop() 
                throws java.util.EmptyStackException {
            stack.pop();
        }
        void newline(StringBuffer sb)
                throws java.util.EmptyStackException {
            int currindent = ((Integer)stack.peek()).intValue();
            sb.append('\n');
            for (int i=0; i<currindent; i++)
                sb.append(' ');
        }
        void newline(Writer w)
                throws java.util.EmptyStackException, java.io.IOException {
            int currindent = ((Integer)stack.peek()).intValue();
            w.write('\n');
            for (int i=0; i<currindent; i++)
                w.write(' ');
        }
    }

    public static void main(String[] args) {
        String inputFile = null;
        String outputFile = null;
        boolean validating = false;
        boolean doTrimWhitespace = true; // if true, trim characters[]
        
        for (int i=0; i<args.length; i++) {
            if (args[i].charAt(0) == '-') {
                if (args[i].startsWith("--output=")) {
                    outputFile = args[i].substring(args[i].indexOf('=')+1);
                } else if (args[i].startsWith("--mode=")) {
                    String m = args[i].substring(args[i].indexOf('=')+1);
                    if (m.equals("xslt"))
                        outputMode = XSLT_MODE;
                    else
                        Usage();
                } else if (args[i].equals("--validating")) {
                    validating = true;
                } else if (args[i].equals("--trim")) {
                    doTrimWhitespace = true;
                } else if (args[i].equals("--notrim")) {
                    doTrimWhitespace = false;
                } else {
                    Usage();
                }
            } else {
                if (inputFile != null)
                    Usage();
                inputFile = args[i];
            }
        }
        
        if (inputFile == null)
            Usage();

        int exitStatus = 1;     // error
        try {

            SAXParserFactory spf = SAXParserFactory.newInstance();
            spf.setValidating(validating);
            spf.setNamespaceAware(true);
            spf.setFeature("http://xml.org/sax/features/namespace-prefixes",
                           false);
            spf.setFeature("http://xml.org/sax/features/namespaces",
                           true);
            XMLReader xr = spf.newSAXParser().getXMLReader();

            LxWriter lxw;
            if (outputFile == null)
                lxw = new LxWriter(new java.io.OutputStreamWriter(System.out));
            else
                lxw = new LxWriter(new File(outputFile));
            lxw.setTrimWhitespace(doTrimWhitespace);
            xr.setContentHandler(lxw);
            xr.setErrorHandler(lxw);

            xr.parse(inputFile);

            exitStatus = 0;     // success!
            
        } catch (javax.xml.parsers.FactoryConfigurationError e) {
            System.err.println("Couldn't create XML parser: " + e);
        } catch (javax.xml.parsers.ParserConfigurationException e) {
            System.err.println("Error configuring parser: " + e);
        } catch (org.xml.sax.SAXException e) {
            System.err.println("Couldn't parse input: " + e);
        } catch (java.io.IOException e) {
            System.err.println("IO error: " + e);
        }

        System.exit(exitStatus);
    }

    private static void Usage() {
        System.err.println("Usage: LxWriter [--validating] [--[no]trim] [--output=outputfile] [--mode=xslt] file.xml");
        System.exit(1);
    }
}
