// This file is part of Lx.
//
// Copyright 2003, 2004, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.reader;

import uk.me.nxg.lx.LxException;


/**
 * Abstracts a set of attributes for an element.
 */
class Attributes
	implements org.xml.sax.Attributes {

    private java.util.List atts;

    public Attributes() {
	atts = new java.util.ArrayList();
    }

    /**
     * Add a new attribute to the set.
     *
     * @param qname the QName of the attribute
     * @param value the value of the attribute
     * @throws LxException if either the name or the value is null
     */
    public void add(QName qname, String value)
            throws LxException {
        if (qname == null || value == null)
            throw new LxException("Neither qname nor value may be null");

        atts.add(new OneAttribute(qname, value));
    }

    /**
     * Adds each of the members of another attribute set to this one.
     *
     * @param newatts the attribute set to be added
     */
    public void add(Attributes newatts) {
        atts.addAll(newatts.atts);
    }
    
    /**
     * Add a namespace mapping attribute.  
     *
     * <p>Note that when SAX2 reports namespace declaration
     * attributes, like <code>xmlns:h</code>, it conforms to the
     * Namespaces in XML recommendation from W3C:  they aren't in any
     * namespace.  Some other W3C specifications violate that
     * specification, and put such declarations into a
     * <code>http://www.w3.org/2000/xmlns/</code> namespace.  This
     * method conforms to the SAX interpretation, and the namespaces
     * attribute is itself in no namespace.
     *
     * @param prefix the prefix being mapped; may be null or the
     * empty string to denote the default namespace
     * @param uri the URI it is being mapped to; must not be null
     * @throws LxException if the URI is null
     */
    public void addNSAttribute(String prefix, String uri)
            throws LxException {
        if (uri == null)
            throw new LxException("Attributes.addNSAttribute(): empty URI");
        atts.add(new OneAttribute(QName.newNSAttribute(prefix), uri));
    }

    /* Implementation of the Attributes interface */

    public int getLength () {
	return atts.size();
    }

    public String getURI (int index) {
	if (index >= atts.size())
	    return null;
        String ns = getatt(index).getNamespace();
 	return ns == null ? "" : ns;
    }

    public String getLocalName (int index) {
	if (index >= atts.size())
	    return null;
        return getatt(index).getLocalName();
    }
    
    public String getQName (int index) {
	if (index >= atts.size())
	    return null;
        return getatt(index).getQName();
    }
    
    public String getType (int index) {
	if (index >= atts.size())
	    return null;
	else
	    return "CDATA";
    }
    
    public String getValue (int index) {
	if (index >= atts.size())
	    return null;
        return getatt(index).getValue();
    }

    public int getIndex (String uri, String localName) {
	if (localName == null || localName.length() == 0)
	    // silly call
	    return -1;

        String testuri = (uri == null || uri.length() == 0
                          ? null
                          : uri.intern());
        String testlocalname = localName.intern();

	for (int i=0; i<atts.size(); i++) {
	    OneAttribute a = getatt(i);
            //	    OneAttribute a = ((OneAttribute)atts.get(i));
	    //String ns = a.getNamespace();
	    if (nsMatch(a, testuri)
		&& a.getLocalName() == testlocalname)
		return i;
	}
	return -1;
    }

    public int getIndex (String qName) {
	if (qName == null || qName.length() == 0)
	    return -1;

        String qntest = qName.intern();

	for (int i=0; i<atts.size(); i++) {
            if (getatt(i).getQName() == qntest)
		return i;
	}
	return -1;
    }

    public String getType (String uri, String localName) {
	if (localName == null || localName.length() == 0)
	    return null;
	
        String testuri = (uri == null || uri.length() == 0
                          ? null
                          : uri.intern());
        String testlocalname = localName.intern();

	for (int i=0; i<atts.size(); i++) {
            OneAttribute a = getatt(i);
	    if (nsMatch(a, testuri)
		&& a.getLocalName() == testlocalname)
		return "CDATA";
	}
	return null;
    }

    public String getType (String qName) {
	if (qName == null || qName.length() == 0)
	    return null;

        String testqname = qName.intern();

	for (int i=0; i<atts.size(); i++) {
            if (getatt(i).getQName() == testqname)
                return "CDATA";
	}
	return null;
    }

    public String getValue (String uri, String localName) {
	if (localName == null || localName.length() == 0)
	    return null;
	
	String testuri = (uri == null || uri.length() == 0
                          ? null
                          : uri.intern());
        String testlocalname = localName.intern();

	for (int i=0; i<atts.size(); i++) {
            OneAttribute a = getatt(i);
	    if (nsMatch(a, testuri)
		&& a.getLocalName() == testlocalname)
		return a.getValue();
	}
	return null;
    }

    public String getValue (String qName) {
	if (qName == null || qName.length() == 0)
	    return null;

        String testqname = qName.intern();
        
	for (int i=0; i<atts.size(); i++) {
            OneAttribute a = getatt(i);
	    if (a.getQName() == testqname)
		return a.getValue();
	}
	return null;
    }

    /* End of org.xml.sax.Attributes implementation */

    /**
     * Clears all the attributes from this set
     */
    public void clear() {
        atts.clear();
    }

    /**
     * Tests if an attribute is present in the attribute set
     * @param qname the QName of an attribute
     * @return true if there is an attribute with this QName in the set
     */
    public boolean isPresent(QName qname) {
        for (int i=0; i<atts.size(); i++)
            if (getatt(i).equalQName(qname))
                return true;
        return false;
    }
    
    /**
     * Tests if an attribute's namespace matches the given URI.
     * @param a an attribute to match
     * @param uri a URI to match it with; if this is non-null, it
     * must have been interned
     */
    private boolean nsMatch(OneAttribute a, String uri) {
	assert a != null;
	String ns = a.getNamespace();
	if (uri != null && uri.length() == 0)
	    uri = null;
	return (uri == null && ns == null)
		|| (uri != null && ns != null && uri == ns);
    }

    /**
     * Returns the attribute with the given index.  If the index is
     * out of range, this throws an assertion error, thus we may assume that,
     * if it returns, it returns successfully.
     * @param index index of the attribute to return
     * @return the attribute at the given index, which is never null
     */
    private OneAttribute getatt(int index) {
        assert index >= 0 && index < atts.size();
        Object o = atts.get(index);
        assert o != null && o instanceof OneAttribute;
        return (OneAttribute)o;
    }

    /**
     * Returns the set of attributes as a string.  Mostly for debugging.
     */
    public String toString() {
        StringBuffer sb=new StringBuffer();
        int len = getLength();
        for (int i=0; i<getLength(); i++)
            sb.append(getQName(i)).append('=').append(getValue(i)).append(';');
        return sb.toString();
    }

    /**
     * Represents a single attribute.
     *
     * <p>The QName, localname, namespace and value of the attribute,
     * as returned by the respective methods of this class, are all
     * interned using {@link java.lang.String#intern}.
     */
    private class OneAttribute {
        private final QName qname;
        private final String value;

        public OneAttribute(QName qname, String value) 
                throws LxException {
            if (qname == null || value == null)
                throw new LxException("OneAttribute has null qname or value");
            this.qname = qname;
            this.value = value.intern();
            assert this.qname != null && this.value != null;
        }

        public String getQName() {
            assert qname != null;
            return qname.qName();
        }

        public String getLocalName() {
            assert qname != null;
            return qname.localName();
        }
        
        public String getNamespace() {
            assert qname != null;
            return qname.namespace();
        }

        public String getValue() {
            assert value != null;
            return value;
        }

        /**
         * Tests whether this attribute has the given qname 
         * @param qname a QName to test
         * @return true if the given QName represents the same
         * namespace-localname pair as this attriubte
         */
        public boolean equalQName(QName qname) {
            return this.qname.equals(qname);
        }

        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append(qname.toString()).append("=").append(getValue());
            return sb.toString();
        }
    }
}

