// This file is part of Lx.
//
// Copyright 2003-2007, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.reader;

import uk.me.nxg.lx.LxException;
import uk.me.nxg.lx.tokenizer.Tokenizer;
import uk.me.nxg.lx.tokenizer.Tokenizer.Token;

import org.xml.sax.ContentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.DTDHandler;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;

import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;

import java.util.Iterator;
import java.io.IOException;
import java.io.Reader;


/**
 * Implements a SAX2 interface on an lx file.
 *
 * <p>The <code>lx</code> syntax is intended to be a compact syntax
 * for XML files, in situations where there is more markup than text
 * content, so that the markup ends up getting in the way.
 *
 * <p>See {@link uk.me.nxg.lx.reader} for a full description of the
 * Lx syntax.
 *
 * @author Norman Gray
 * @version 0.5.3, 2012 August 5
 */
public class LxReader
	implements org.xml.sax.XMLReader {

    /* XXX NOTE: if the namespace strings, or the defaults, are
     * changed, then there must be parallel changes in the
     * documentation in package.html. */

    boolean verbose = false;

    /** 
     * The namespace prefix for all the Lx features.  This has the
     * value <code>http://ns.nxg.me.uk/lx</code>
     */
    public static final String LXNAMESPACE = "http://ns.nxg.me.uk/lx";

    /** The default prefix for Lx elements */
    private static final String LXDEFAULTPREFIX = "lx";

    /**
     * The Lx feature which controls whether attributes inherit the
     * namespace of the element they are contained within.  If this
     * is <em>false</em>, then attributes without a prefix
     * obey the normal rules, so that they if they have no prefix,
     * then they are in no namespace, or else in the namespace to
     * which the null prefix is mapped, if it is.  If this feature is
     * <em>true</em>, then attributes without a prefix acquire the same prefix
     * and namespace as that of the element they are contained within.
     *
     * <p>The latter behaviour -- value true -- is probably the more
     * useful, but the default is false so that the behaviour matches
     * that of XML.
     *
     * <p>This feature string has the value
     * <code>http://ns.nxg.me.uk/lx/attributes-inherit-namespace</code>,
     * and the default value is <code>false</code>.
     *
     * <p>It is illegal to set both this feature and {@link
     * #LX_ATTRIBUTES_IN_NO_NAMESPACE} to be true.
     */
    public static final String LX_ATTRIBUTES_INHERIT_NAMESPACE
            = "http://ns.nxg.me.uk/lx/attributes-inherit-namespace";

    /**
     * The Lx feature which controls whether unprefixed attributes end
     * up in no namespace.  If this is <em>false</em>, then attributes
     * without a prefix obey the normal rules.  If it is
     * <em>true</em>, then attributes without a namespace are always
     * put into the empty namespace, even if the null prefix is
     * currently mapped to a non-null namespace.
     *
     * <p>The latter behaviour is both inconsistent and rather
     * counterintuitive; however it is the default in order to make
     * this class's default behaviour consistent with XML.
     *
     * <p>This feature string has the value
     * <code>http://ns.nxg.me.uk/lx/attributes-in-no-namespace</code>,
     * and has the default value <code>true</code>.
     *
     * <p>It is illegal to set both this feature and {@link
     * #LX_ATTRIBUTES_INHERIT_NAMESPACE} to be true.
     */
    public static final String LX_ATTRIBUTES_IN_NO_NAMESPACE
            = "http://ns.nxg.me.uk/lx/attributes-in-no-namespace";

    /**
     * This property contains the string which is used when generating
     * prefixes.  The feature URL has the string value
     * <code>http://ns.nxg.me.uk/lx/generated-prefix</code>, and the
     * default value <code>lxns</code>.
     */
    public static final String LX_GENERATED_PREFIX
            = "http://ns.nxg.me.uk/lx/generated-prefix";

    /**
     * The SAX feature which controls whether namespace information is
     * reported.  If it is reported, it is done through the
     * {@link org.xml.sax.ContentHandler#startElement} and 
     * {@link org.xml.sax.ContentHandler#endElement} callbacks.  This SAX
     * implementation supports both the values <code>true</code> (the
     * default) and <code>false</code> (in fact, its behaviour at
     * present is identical in both cases).
     *
     * <p>See the <a
     * href="http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description"
     * >SAX documentation</a> for detailed discussion of SAX features.
     */
   public static final String SAX_NAMESPACES
           = "http://xml.org/sax/features/namespaces";

    /**
     * The SAX feature which controls whether namespace mapping
     * attributes are reported.
     *
     * <p>See the <a
     * href="http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description"
     * >SAX documentation</a> for detailed discussion of SAX features.
     */
    public static final String SAX_NAMESPACE_PREFIXES
            = "http://xml.org/sax/features/namespace-prefixes";

    /**
     * The SAX feature which indicates whether all XML names have been
     * interned.
     *
     * <p>See the <a
     * href="http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description"
     * >SAX documentation</a> for detailed discussion of SAX features.
     */
    public static final String SAX_STRINGS_INTERNED
            = "http://xml.org/sax/features/string-interning";

    private ContentHandler contenthandler;
    private ErrorHandler errorhandler;
    private DTDHandler dtdhandler;
    private EntityResolver entityresolver;

    private Stack elementStack;
    /**
     * A collection of namespace attributes, accumulated before the
     * current element.  Used both for supporting SAX
     * namespace-prefixes reporting and for preparing the set of
     * endPrefixMapping calls.
     */
    private Attributes nsAttributes;
    private java.util.Map featureMap;

    /**
     * Creates a new reader
     *
     * @throws LxException if there is any problem creating the reader
     */
    public LxReader() 
            throws LxException {
        
        elementStack = new Stack();
        nsAttributes = new Attributes();

        featureMap = new java.util.HashMap();
        featureMap.put(SAX_NAMESPACES,
                       new OneFeature(SAX_NAMESPACES, true, true));
        featureMap.put(SAX_NAMESPACE_PREFIXES,
                       new OneFeature(SAX_NAMESPACE_PREFIXES, false, true));
        featureMap.put(SAX_STRINGS_INTERNED,
                       new OneFeature(SAX_STRINGS_INTERNED, true, false));
        featureMap.put(LX_ATTRIBUTES_INHERIT_NAMESPACE,
                       new OneFeature(LX_ATTRIBUTES_INHERIT_NAMESPACE,
                                      false, true));
        featureMap.put(LX_ATTRIBUTES_IN_NO_NAMESPACE,
                       new OneFeature(LX_ATTRIBUTES_IN_NO_NAMESPACE,
                                      true, true));
        // property LX_GENERATED_PREFIX is also initialised in
        // uk.me.nxg.reader.Stack, but has to be initialised here, too,
        // in order to insert this in the list of recognised properties/features.
        featureMap.put(LX_GENERATED_PREFIX,
                       new OneProperty(LX_GENERATED_PREFIX, "lxns", true));
        for (java.util.Iterator mi = featureMap.keySet().iterator();
             mi.hasNext(); ) {
            String k = (String)mi.next();
            // The stack is empty, so these are set in the base map
            Object o = featureMap.get(k);
            if (o instanceof OneFeature)
                elementStack.setFeature(k, ((OneFeature)o).value);
            else
                elementStack.setProperty(k, ((OneProperty)o).value);
        }
        
        elementStack.setPrefixMapping(LXDEFAULTPREFIX, LXNAMESPACE);
    }

    /**
     * Switch on or off debugging output.
     * @param verbose true if copious debugging output is to be produced
     */
    public void setVerbose(boolean verbose) {
        this.verbose = verbose;
    }

    /* Implementation of XMLReader interface */

    public boolean getFeature (String name)
	    throws SAXNotRecognizedException, SAXNotSupportedException {
        if (recognisedFeature(name))
            return elementStack.getFeature(name);
        // else
        throw new SAXNotRecognizedException("unrecognised feature: " + name);
    }

    private boolean recognisedFeature(String name) {
        return featureMap.containsKey(name);
    }

    public void setFeature (String name, boolean value)
	    throws SAXNotRecognizedException, SAXNotSupportedException {
        name = name.intern();
        OneFeature f = (OneFeature)featureMap.get(name);
        if (f == null)
            throw new SAXNotRecognizedException
                    ("Unrecognised feature " + name);

        boolean currentValue = elementStack.getFeature(name);
        if (value == currentValue)
            return;             // do nothing, and don't test mutability

        if (f.mutable) {
            elementStack.setFeature(name, value);
        } else {
            throw new SAXNotSupportedException
                    ("Cannot set feature " + name + " to non-default value");
        }
        
	return;
    }

    public Object getProperty (String name)
	    throws SAXNotRecognizedException, SAXNotSupportedException {
        if (recognisedFeature(name))
            return elementStack.getProperty(name);
        // else
	throw new SAXNotRecognizedException
		("Property " + name + " not recognised");
    }

    public void setProperty (String name, Object value)
	    throws SAXNotRecognizedException, SAXNotSupportedException {
        name = name.intern();
        OneProperty p = (OneProperty)featureMap.get(name);
        if (p == null)
            throw new SAXNotRecognizedException
                    ("Unrecognised property " + name);

        Object currentValue = elementStack.getProperty(name);
        if (value == currentValue)
            return;             // do nothing, and don't test mutability

        if (p.mutable) {
            elementStack.setProperty(name, value);
        } else {
            throw new SAXNotSupportedException
                    ("Cannot set property " + name + " to non-default value");
        }
        
	return;
    }
	
    public void setEntityResolver (EntityResolver resolver) {
	if (resolver == null)
	    throw new NullPointerException("setEntityResolver: null resolver");
	entityresolver = resolver;
    }
    
    public EntityResolver getEntityResolver () {
	return entityresolver;
    }
    
    public void setDTDHandler (DTDHandler handler) {
	if (handler == null)
	    throw new NullPointerException("setDTDHandler: null handler");
	dtdhandler = handler;
    }
    
    public DTDHandler getDTDHandler () {
	return dtdhandler;
    }
    
    public void setContentHandler (ContentHandler handler) {
	if (handler == null)
	    throw new NullPointerException("setContentHandler: null handler");
	contenthandler = handler;
    }
    
    public ContentHandler getContentHandler () {
	return contenthandler;
    }
    
    public void setErrorHandler (ErrorHandler handler) {
	if (handler == null)
	    throw new NullPointerException("setErrorHandler: null handler");
	errorhandler = handler;
    }
    
    public ErrorHandler getErrorHandler () {
	return errorhandler;
    }
    
    public void parse(String systemId)
	    throws IOException, SAXException {
	parse(new InputSource(systemId));
    }

    public void parse(InputSource input)
	    throws IOException, SAXException {
        if (input == null)
            throw new IOException("asked to parse a null <input>");
        Reader r;

        r = input.getCharacterStream();
        if (r == null) {
            java.io.InputStream is = input.getByteStream();
            try {
                if (is != null) {
                    String encoding = input.getEncoding();
                    if (encoding == null)
                        r = new java.io.InputStreamReader(is);
                    else
                        r = new java.io.InputStreamReader(is, encoding);
                }
            } catch (java.io.UnsupportedEncodingException e) {
                throw new IOException
                        ("Unsupported encoding in InputSource: " + e);
                // ...or just ignore the encoding?
            }
        }
        if (r == null) {
            String sysid = input.getSystemId();
            if (sysid != null) {
                // FIXME: should handle (or at least test for) cases where
                // the sysid isn't a URL that this can handle, but this and
                // the following fixme will probably need to wait until I
                // have a concrete case that fails
                java.net.URL url = new java.net.URL(sysid);
                // FIXME: should examine the URL's content-encoding
                r = new java.io.InputStreamReader(url.openStream());
            }
        }
        if (r == null)
            throw new IOException("Unable to get content of InputSource");

	Tokenizer t = new Tokenizer(r);
	t.setSystemId(input.getSystemId());
        parse(t);
    }

    /**
     * Start to parse using the given Tokenizer.  This is an
     * extension to the SAX interface.  Generally, you will wish to
     * use either of the standard SAX parse methods, but if you have
     * some reason to create a different or extended implementation of the
     * Tokenizer class, you can invoke a parse on it with this method.
     *
     * @param tokenizer an object which will return tokens from a
     * source
     * @see #parse(String)
     * @see #parse(InputSource)
     */
    public void parse(Tokenizer tokenizer)
	    throws IOException, SAXException {
	// The business!
        if (tokenizer == null)
            throw new IOException("asked to parse a null <Tokenizer>");

        Tokenizer t = tokenizer;

        assert elementStack.empty();

        Token nextToken;
	Element el;

	if (contenthandler == null)
	    throw new SAXParseException
		    ("parsing started with no ContentHandler registered", t);
	
	contenthandler.setDocumentLocator(t);

        if (verbose)
            System.err.println("startDocument");
	contenthandler.startDocument();

	try {
	    while ((nextToken = t.getToken()) != Token.EOF) {
		switch (nextToken) {
		  case OPENPAREN:
		    try {
			t.ungetToken(nextToken);
			el = new Element(t, elementStack);
		    } catch (LxException e) {
			throw new SAXParseException
				("Error parsing stream: " + e, t, e);
		    }

                    assert contenthandler != null;

                    if (el.namespace() == LXNAMESPACE) {

                        handleLxCommand(el.localName(), t, contenthandler);

                    } else {

                        if (nsAttributes.getLength() > 0) {
                            el.atts().add(nsAttributes);
                            nsAttributes.clear();
                        }

                        if (verbose)
                            System.err.println("start: {" + el.namespace()
                                               + "}" + el.localName()
                                               + "(" + el.qName() + ") "
                                               + el.atts());

                        contenthandler.startElement(el.namespace(),
                                                    el.localName(),
                                                    el.qName(),
                                                    el.atts());
                        final String ns = el.namespace();
                        final String ln = el.localName();
                        final String qn = el.qName();
                        Stack.Promise np
                                = new Stack.Promise() {
                                        final ContentHandler ch
                                                = contenthandler;
                                        protected void doit() {
                                            try {
                                                if (verbose)
                                                    System.err.println("end: " + qn);
                                                ch.endElement(ns, ln, qn);
                                            } catch (SAXException e) {
                                                // ignore it
                                            }
                                        }
                                    };
                        elementStack.push(new Stack.Element(el.qName(), np));
                    }
		    break;

		  case CLOSEPAREN:
		    if (elementStack.empty()) {
			throw new SAXParseException
				("Too many right parentheses (line "
                                 + t.getLineNumber() + ")", t);
		    }

                    Stack.Promise p = elementStack.pop();
                    if (p != null)
                        p.force();
		    break;

		  case SYMBOL: // distinguish these later
		  case STRING:
		    String val = t.value();
                    if (verbose)
                        System.err.println("chars: " + val);
		    contenthandler.characters
			    (val.toCharArray(), 0, val.length());
		    break;

		  case TOKEN:
                    throw new SAXParseException
                            ("Token [#" + t.value()
                             + "] out of context in input stream (line "
                             + t.getLineNumber() + ")",
                             t);

		  case KEYWORD:
                    throw new SAXParseException
                            ("Keyword [" + t.value()
                             + ":] out of context in input stream (line "
                             + t.getLineNumber() + ")"
                             , t);

		  default:
		    assert(nextToken != Token.EOF);
		    throw new SAXParseException
			    ("Unrecognised token in input stream (line "
                             + t.getLineNumber() + ")", t);
		}
	    }
	} catch (LxException e) {
	    throw new SAXParseException("Tokenization exception (line "
                                        + t.getLineNumber() + "): "+e, t, e);
	}

	if (! elementStack.empty()) {
	    throw new SAXParseException("Too many left parentheses", t);
	}

        if (verbose)
            System.err.println("endDocument");
	contenthandler.endDocument();
    }

    /* End of implementation of XMLReader interface */

    /**
     * Handle the given command.
     *
     * @param cmd the command which was read from the input
     * @param t the Tokenizer which we may read arguments from,
     * positioned immediately after the command which triggered this call
     * @param ch the ContentHandler we should send events to, if that
     * is appropriate
     * @throws SAXParseException if there is a parse or IO error
     * reading the stream
     * @throws SAXNotRecognizedException if a <code>lx:feature</code>
     * command attempts to set an unrecognised feature
     */                        
    private void handleLxCommand(String cmd,
                                 Tokenizer t,
                                 final ContentHandler ch) 
            throws SAXParseException, SAXNotRecognizedException {

        // constants
        final int lxnamespace = 1;
        final int lxdtd = 2;
        final int lxfeature = 3;
        final int lxproperty = 4;
        final int lxlet = 5;
        int whichcommand = 0;

        if (cmd.equals("namespace"))
            whichcommand = lxnamespace;
        else if (cmd.equals("dtd"))
            whichcommand = lxdtd;
        else if (cmd.equals("feature"))
            whichcommand = lxfeature;
        else if (cmd.equals("property"))
            whichcommand = lxproperty;
        else if (cmd.equals("let"))
            whichcommand = lxlet;
        else
            throw new SAXParseException
                    ("Unexpected Lx command " + cmd, t);

        // command is OK

        // Now read an object from the stream; it should be a list
        // for passing to the real command handler
        LxObject obj;
        try {
            obj = LxObject.readLxObject(t);
            if (!obj.isList()) {
                throw new SAXParseException
                        ("Malformed declaration: expected list, got " + obj,
                         t);
            }
        } catch (LxException e) {
            throw new SAXParseException("Error reading input: " + e, t, e);
        }


        Stack.Mappings map = elementStack.newMapping();
        elementStack.push(map);

        switch (whichcommand) {
          case lxnamespace:
            handleNamespaceDeclaration((LxList)obj, ch, map, t);
            break;
          case lxdtd:
            handleDtdDeclaration((LxList)obj, ch, map, t);
            break;
          case lxfeature:
            handleFeatureDeclaration((LxList)obj, ch, map, t);
            break;
          case lxproperty:
            handlePropertyDeclaration((LxList)obj, ch, map, t);
            break;
          case lxlet:
            handleLetDeclaration((LxList)obj, ch, map, t);
            break;
          default:
            assert false;
        }

        return;
    }

    /**
     * Handle a namespace declaration.  The Tokenizer is positioned
     * just before the list of lists which is the first argument to
     * the namespace declaration
     *
     * @param nslist the list of namespace declarations
     * @param ch the ContentHandler for this parse
     * @param pmap the Mappings object we should add mappings to
     * @param loc a Locator, used for reporting parse errors
     */
    private void handleNamespaceDeclaration(LxList nslist,
                                            final ContentHandler ch,
                                            final Stack.Mappings pmap,
                                            Locator loc)
            throws SAXParseException {

        try {
            // nslist should be a list of lists.
            // Each of those lists is of the form (prefix uri).
            for (;
                 nslist != LxList.EMPTYLIST;
                 nslist = nslist.cdr()) {
                if (! nslist.car().isList())
                    throw new SAXParseException
                            ("Non-list in namespace declaration: " 
                             + nslist.car(),
                             loc);
                
                LxList nsmap = (LxList)nslist.car();
                if (nsmap.length() != 2)
                    throw new SAXParseException
                            ("Bad element in namespace declaration: " + nsmap,
                             loc);
                LxObject prefix = nsmap.car();
                LxObject uri = nsmap.cdr().car();
                String ps;
                String uris;
                
                try {
                    // validate type of prefix, and convert to string in ps
                    if (prefix.isTrue()) {
                        if (prefix.isSymbol() || prefix.isString()) {
                            ps = prefix.toString();
                        } else {
                            throw new SAXParseException
                                    ("Prefix <" + prefix.toString()
                                     + "> of wrong type", loc);
                        }
                    } else {
                        // prefix is #f
                        ps = null;
                    }

                    // same for URI
                    if (uri.isTrue()) {
                        if (uri.isString()) {
                            uris = uri.toString();
                            if (uris.length() == 0)
                                uris = null;
                        } else {
                            throw new SAXParseException
                                    ("Mapping prefix <" + ps
                                     + "> to object of wrong type: "
                                     + uri.toString(),
                                     loc);
                        }
                    } else {
                        uris = null;
                    }

                    if (ps == null && uris == null) {
                        // (#t #t) or (#t ""): legal no-op
                    } else {
                        pmap.mapPrefix(ps, uris);
                    }
                    
                } catch (SAXException e) {
                    // convert it to a parsing exception
                    throw new SAXParseException
                            ("Unexpected SAX exception: " + e, loc);
                }
            }

            // Create a Promise which will invoke the endPrefixMapping
            // handlers in the opposite order to the order we'll
            // invoke the startPrefixMapping handlers below.  This
            // isn't mandated by the SAX interface, but it makes
            // testing easier for us, and generally makes life easier for
            // the eventual clients.  This relies on Stack.Mappings
            // to return an iterator which has a predictable order
            // (which happens to be insertion order).
            //
            // Order matters here -- we must call
            // {start,end}PrefixMapping using the possibly generated
            // prefixes that are in pmap, and the only way we have of
            // getting to them is via elementStack.getPrefixMapping,
            // and that will only work _after_ we have pushed pmap
            // onto the stack.  Also, the Promise will be executed
            // _after_ pmap has been removed from the stack, so we
            // must make these getPrefixMapping calls before we
            // create the Promise object, and not within it.
            {
                // Create a list of all the prefixes in the opposite
                // order to that (stable) order in which they're
                // retrieved from the pmap iterator.
                final java.util.List prefixList = new java.util.LinkedList();
                for (Iterator mi = pmap.getKeyIterator(); mi.hasNext();  ) {
                    Stack.PrefixNamespacePair pnp
                            = elementStack.getPrefixMapping((String)mi.next());
                    assert pnp != null;
                    if (pnp.getNamespace() != null)
                        prefixList.add(0, pnp.getPrefix());
                }

                pmap.addPromise(new Stack.Promise() {
                        protected void doit() {
                            try {
                                for (Iterator li = prefixList.iterator();
                                     li.hasNext(); )
                                    ch.endPrefixMapping((String)li.next());
                            } catch (SAXException e) {
                                // Can't happen
                                assert false;
                            }
                        }
                    });
            }
            
            for (Iterator mi = pmap.getKeyIterator(); mi.hasNext(); ) {
                String p = (String)mi.next();
                Stack.PrefixNamespacePair pnp
                        = elementStack.getPrefixMapping(p);
                assert pnp != null;
                String uri = pnp.getNamespace();
                if (uri != null) {
                    ch.startPrefixMapping(pnp.getPrefix(), pnp.getNamespace());
                    if (elementStack.getFeature(SAX_NAMESPACE_PREFIXES))
                        nsAttributes.addNSAttribute(pnp.getPrefix(),
                                                    pnp.getNamespace());
                }
            }
            
        } catch (SAXNotRecognizedException e) {
            // this is from getFeature, and can't happen!
            throw new AssertionError
                    ("Can't happen: SAX feature " + SAX_NAMESPACE_PREFIXES
                     + " wasn't recognised: " + e);
        } catch (SAXException e) {
            // this is from startPrefixMapping
            throw new SAXParseException("Error mapping prefix: " + e, loc);
        } catch (LxException e) {
            throw new SAXParseException
                    ("Error reading from stream: " + e, loc, e);
        }
    }

    /**
     * Handles a DTD declaration.  Syntax is
     * <pre>
     * (lx:dtd ((el1 (att1-1 val1-1)
     *               (att1-2 val1-2)
     *                ...)
     *          (el2 (att2-1 val2-1)
     *               ...))
     *          ...))
     *    (body)
     *    ...)
     * </pre>
     * 
     * The <code>att*</code> are attribute names.  An element may
     * have only the attributes that are declared, and if
     * <code>val*</code> is true, the attribute must be present.  If
     * there is more than one attribute which must be present they
     * must appear in the given order.
     *
     * <p>If the attribute list is replaced by a <code>#f</code>, as
     * in <code>(el1 #f)</code>, then the given element has
     * unrestricted content.  It is an error for this token
     * <code>#f</code> to appear in any other position.
     *
     * @param elementlist the list of element declarations
     * @param ch the ContentHandler for this parse
     * @param elmap the Mappings object we should add mappings to
     * @param loc a Locator, used for reporting parse errors
     */
    private void handleDtdDeclaration(LxList elementlist,
                                      final ContentHandler ch,
                                      final Stack.Mappings elmap,
                                      Locator loc)
            throws SAXParseException {
        try {
            // elementlist should be a list of lists.
            // Each of those lists is of the form given above
            for (;
                 elementlist != LxList.EMPTYLIST;
                 elementlist = elementlist.cdr()) {
                if (! elementlist.car().isList())
                    throw new SAXParseException
                            ("Non-list in dtd declaration: "
                             + elementlist.car(),
                             loc);
                LxList eldecn = (LxList)elementlist.car();
                // eldecn = (el (att val) (att) "att" ...)
                if (eldecn.car().isStringLike()) {
                    elmap.mapElementDecl(new QName(eldecn.car().toString(),
                                                   elementStack),
                                         eldecn.cdr());
                } else {
                    throw new SAXParseException
                            ("dtd declaration: malformed element list: "
                             + eldecn, loc);
                }
            }

         } catch (LxException e) {
             throw new SAXParseException("Error reading input: " + e, loc, e);
         }
    }

    /**
     * Handles a feature declaration.  Syntax is
     * <pre>
     * (lx:feature (&lt;feature-spec&gt; ...)
     *    ..)
     * </pre>
     * Where <code>&lt;feature-spec&gt;</code> is either
     * <code>("feature-string" boolean)</code>, indicating that the
     * feature with the URI given in the string should have the
     * given boolean value; or else just <code>"feature-string"</code>,
     * in which case the given feature is set true.
     *
     * @param featlist a list of feature declarations
     * @param ch a content handler which is to receive any generated events
     * @param featmap the Mappings object we should add mappings to
     * @param loc a Locator, used for reporting parse errors
     * @throws SAXParseException if we find illegal syntax
     * @throws SAXNotRecognizedException if we are asked to set an
     * unknown feature
     */
    private void handleFeatureDeclaration(LxList featlist,
                                          final ContentHandler ch,
                                          final Stack.Mappings featmap,
                                          Locator loc)
            throws SAXParseException, SAXNotRecognizedException {

        for ( ;
              featlist != LxList.EMPTYLIST;
              featlist = featlist.cdr()) {

            LxObject car = featlist.car();
            String featurename = null;
            boolean featurevalue = false;
                
            if (car.isList()) {

                LxList featpair = (LxList)car;
                if (featpair.length() != 2)
                    throw new SAXParseException
                            ("Bad element in feature declaration: "
                             + featpair,
                             loc);
                LxObject featureURI = featpair.car();
                LxObject value = featpair.cdr().car();
                // Validate type of feature
                if (featureURI.isString()) {
                    featurename = featureURI.toString();
                    featurevalue = value.isTrue();
                    
                } else {
                    throw new SAXParseException
                            ("bad type for feature: " + featureURI, loc);
                }

            } else if (car.isString()) {

                featurename = car.toString();
                featurevalue = true;
                    
            } else {
                    
                throw new SAXParseException
                        ("Bad type in feature declaration: " + car, loc);
            }

            if (recognisedFeature(featurename))
                featmap.mapFeature(featurename, featurevalue);
            else
                throw new SAXNotRecognizedException("Feature "
                                                    + featurename
                                                    + " not recognised");
                
        }
    }

    /**
     * Handles a property declaration.  Syntax is
     * <pre>
     * (lx:property ((&lt;property-name&gt; &lt;property-value&gt;) ...)
     *    ..)
     * </pre>
     * where <code>&lt;property-name&gt;</code> is one of the properties
     * available.
     *
     * @param proplist a list of property declarations
     * @param ch a content handler which is to receive any generated events
     * @param featmap the Mappings object we should add mappings to
     * @param loc a Locator, used for reporting parse errors
     * @throws SAXParseException if we find illegal syntax
     * @throws SAXNotRecognizedException if we are asked to set an
     * unknown feature
     */
    private void handlePropertyDeclaration(LxList proplist,
                                           final ContentHandler ch,
                                           final Stack.Mappings propmap,
                                           Locator loc)
            throws SAXParseException, SAXNotRecognizedException {

        for ( ;
              proplist != LxList.EMPTYLIST;
              proplist = proplist.cdr()) {

            LxObject car = proplist.car();
            String propertyname = null;
            String propertyvalue = null;
                
            if (car.isList()) {

                LxList proppair = (LxList)car;
                if (proppair.length() != 2)
                    throw new SAXParseException
                            ("Bad element in property declaration: "
                             + proppair,
                             loc);
                LxObject propertyURI = proppair.car();
                LxObject value = proppair.cdr().car();
                // Validate type of property
                if (propertyURI.isString()) {
                    propertyname = propertyURI.toString();
                    propertyvalue = value.toString();
                    
                } else {
                    throw new SAXParseException
                            ("bad type for property: " + propertyURI, loc);
                }
                    
            } else {
                    
                throw new SAXParseException
                        ("Bad type in property declaration: " + car, loc);
            }

            if (recognisedFeature(propertyname))
                propmap.mapProperty(propertyname, propertyvalue);
            else
                throw new SAXNotRecognizedException("Property "
                                                    + propertyname
                                                    + " not recognised");
        }
    }

    /**
     * Handle a lx:let declaration.
     *
     * <p>Syntax
     * <pre>
     * let-decl := (lx:let (&lt;let-spec&gt; ...) &lt;body&gt;)
     * let-spec := (&lt;command&gt; list ...)
     * command  := "namespace" | "dtd" | "feature" | "property"
     * </pre>
     * Note that the sytax of the <code>&lt;let-decl&gt;</code> is
     * slightly different from that of the corresponding lx command.
     *
     * @param letlist a list of commands
     * @param ch a content handler which is to receive any generated events
     * @param map the Mappings object we should add mappings to
     * @param loc a Locator, used for reporting parse errors
     * @throws SAXParseException if we find illegal syntax
     * @throws SAXNotRecognizedException if we are asked to set an
     * unknown feature
     */
    private void handleLetDeclaration(LxList letlist,
                                      final ContentHandler ch,
                                      final Stack.Mappings map,
                                      Locator loc)
            throws SAXParseException, SAXNotRecognizedException {

        // letlist should be a list of lists;
        // each one is of the form (cmd list)
        for ( ;
              letlist != LxList.EMPTYLIST;
              letlist = letlist.cdr()) {

            LxObject car = letlist.car();
            if (!car.isList()) {
                throw new SAXParseException
                        ("Malformed lx:let declaration: expected list, got "
                         + car, loc);
            }

            LxList letpair = (LxList)car;
            if (! (letpair.car().isSymbol() || letpair.car().isString()))
                throw new SAXParseException
                        ("Bad element in let declaration: expected string, got "
                         + letpair.car(), loc);

            String cmd = letpair.car().toString();
            LxList arglist = letpair.cdr();

            if (cmd.equals("namespace"))
                handleNamespaceDeclaration(arglist, ch, map, loc);
            else if (cmd.equals("dtd"))
                handleDtdDeclaration(arglist, ch, map, loc);
            else if (cmd.equals("feature"))
                handleFeatureDeclaration(arglist, ch, map, loc);
            else if (cmd.equals("property"))
                handlePropertyDeclaration(arglist, ch, map, loc);
            else
                assert false;
        }
    }
    
    private static class OneFeature {
        String name;
        boolean value;
        boolean mutable;
        OneFeature(String n, boolean v, boolean m) {
            name = n;
            value = v;
            mutable = m;
        }
    }

    private static class OneProperty {
        String name;
        Object value;
        boolean mutable;
        OneProperty(String n, Object v, boolean m) {
            name = n;
            value = v;
            mutable = m;
        }
    }
}

