// This file is part of Lx.
//
// Copyright 2003, 2004, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.reader;

import uk.me.nxg.lx.LxException;

/**
 * Represents a qName.
 *
 * <p>The localname, prefix, QName and URI
 * returned by the respective methods all return Strings which have
 * been interned with {@link java.lang.String#intern}.
 */
class QName {
    private String localname;
    private String prefix;
    private String qName;
    private String uri;

    /**
     * Pattern for a QName.
     *
     * <p>See <code>http://www.w3.org/TR/REC-xml-names/</code> for
     * the syntax (reproduced below).  We do not at present support
     * <code>CombiningChar</code> or <code>Extender</code> characters
     * in the QName.
     *
     * <p>We <em>extend</em> the syntax, however, by permitting a
     * null prefix.
     *
     * <pre>
     * [4] NCName ::= (Letter | '_') ( NCNameChar )* 
     * [5] NCNameChar ::= Letter | Digit | '.' | '-' | '_' | 
     *                    CombiningChar | Extender 
     * [6] QName ::= (Prefix ':')? LocalPart 
     * [7] Prefix ::= NCName
     * [8] LocalPart ::= NCName 
     * </pre>
     */
    private static final java.util.regex.Pattern qnamePattern
            = java.util.regex.Pattern.compile
            ("^(([a-zA-Z_][a-zA-Z0-9._-]*)?:)?[a-zA-Z_][a-zA-Z0-9._-]*$");

    /**
     * Pattern for a NSAttName.  Much as above.  For the
     * NSAttName syntax, see
     * <code>http://www.w3.org/TR/REC-xml-names/</code>.
     * <pre>
     * [1] NSAttName ::= PrefixedAttName | DefaultAttName 
     * [2] PrefixedAttName ::= 'xmlns:' NCName 
     * [3] DefaultAttName ::= 'xmlns' 
     * </pre>
     */
    private static final java.util.regex.Pattern nsAttNamePattern
            = java.util.regex.Pattern.compile
            ("^xmlns(:[a-zA-Z_][a-zA-Z0-9._-]*)?$");
    

    /**
     * Creates a new QName, by parsing the GI string passed to it.
     * The prefix, or lack of one, in this string is resolved in the
     * context of the Stack which is also provided.
     *
     * <p>It is in this object that we use the information about
     * namespaces which has been built up from the various
     * <code>lx:namespace</code> declarations.  We map the null (that
     * is, <code>#f</code>), empty (<code>""</code>) or
     * non-empty prefixes to prefix-namespace pairs, and store the
     * latter.  The namespaces may be either null or non-empty -- the
     * empty namespace <code>""</code> is earlier parsed as
     * <code>null</code>.  The legal combinations are as follows:
     *
     * <table>
     * <tr><th>case</th>
     *     <th>input</th><th>namespace</th><th>declaration</th><th>result</th>
     * </tr>
     *
     * <tr><td>1</td>
     *     <td>gi  </td><td>null</td><td>-         </td><td>gi      </td></tr>
     * <tr><td>2</td>
     *     <td>gi  </td><td>ns  </td><td>(#f "ns") </td><td>{ns}G:gi</td></tr>
     * <tr><td>3</td>
     *     <td>:gi </td><td>null</td><td>("" #f)   </td><td>gi      </td></tr>
     * <tr><td>4</td>
     *     <td>:gi </td><td>ns  </td><td>("" "ns") </td><td>{ns}G:gi</td></tr>
     * <tr><td>5</td>
     *     <td>x:gi</td><td>null</td><td>("x" #f)  </td><td>gi      </td></tr>
     * <tr><td>6</td>
     *     <td>x:gi</td><td>ns  </td><td>("x" "ns")</td><td>{ns}x:gi</td></tr>
     *
     * </table>
     *
     * The first case is the default, and would be equivalent to
     * <code>(#f #f)</code>, which is permitted syntactically, but
     * acts as a no-op.  The <code>G</code> indicates a generated
     * prefix.
     *
     * <p>If the third argument, <code>contextQName</code> is
     * non-null, then the constructor examines the values of the
     * {@link LxReader#LX_ATTRIBUTES_INHERIT_NAMESPACE} and
     * {@link LxReader#LX_ATTRIBUTES_IN_NO_NAMESPACE} features, and
     * determines the prefix for the QName based on these and the
     * <code>contextQName</code> (in fact, the
     * <code>contextQName</code> is examined only in the former case).
     * The <code>contextQName</code> should be provided for all
     * attribute QNames, and not for element QNames.  Note that this
     * does not affect QNames with a non-null prefix which maps to the
     * empty namespace, but does affect QNames with a null prefix
     * which maps to a non-null namespace.
     *
     * @param qnameString a non-null QName which is to be parsed
     * @param stack an element/namespace stack which provides context
     * for features and namespace mappings; this must not be null
     * @param contextQName a context QName (typically the element name
     * if this is an attribute QName), used when the
     * <code>attributes-inherit-namespace</code> feature is true
     * @throws LxException if there is an error parsing the string,
     * or if the QName was syntactically invalid
     * @throws IllegalArgumentException if either qnameString or stack is null
     */
    public QName(String qnameString, Stack stack, QName contextQName)
	    throws LxException {

        if (qnameString == null || stack == null)
            throw new IllegalArgumentException
                    ("QName constructor: qnameString=" + qnameString
                     + " and stack must not be null");

        if (! qnamePattern.matcher(qnameString).matches())
            throw new LxException("Malformed QName: " + qnameString);

        String[] components = qnameString.split(":");
        String lxprefix = null;
        switch (components.length) {
          case 0:
            assert false;
            break;

          case 1:
            lxprefix = null;
            localname = components[0];
            break;
                
          case 2:
            lxprefix = components[0];
            localname = components[1];
            break;

          default:
            assert false;       // pattern matching test above makes
                                // this case impossible
        }

        localname = localname.intern();

        if (lxprefix == null) {
            // Is either of the LX_ATTRIBUTES features set?
            // They shouldn't both be set.
            if (contextQName != null
                && stack.getFeature
                (LxReader.LX_ATTRIBUTES_INHERIT_NAMESPACE)) {

                prefix = contextQName.prefix();
                uri = contextQName.namespace();
                qName = (prefix == null
                         ? localname
                         : (prefix + ":" + localname).intern());

            } else if (contextQName != null
                       && stack.getFeature
                       (LxReader.LX_ATTRIBUTES_IN_NO_NAMESPACE)) {

                prefix = null;
                uri = null;
                qName = localname;

            } else {

                // the normal case
                Stack.PrefixNamespacePair pnp = stack.getPrefixMapping(null);
                // null prefix may be mapped to a generated prefix name
                if (pnp == null) {
                    // (case 1)
                    uri = null; // it's not -- that's OK
                    prefix = null;
                    qName = localname;
                } else {
                    // (case 2)
                    prefix = pnp.getPrefix(); // already interned
                    uri = pnp.getNamespace(); // already interned
                    assert uri != null;
                    qName = (prefix + ":" + localname).intern();
                }
            }
        } else {
            assert lxprefix != null;
            Stack.PrefixNamespacePair pnp = stack.getPrefixMapping(lxprefix);
            if (pnp == null) {
                // That's bad -- a non-null prefix should always
                // be mapped.  It may be mapped to a null
                // namespace, but it should still appear in this map.
                throw new LxException("Prefix <" + lxprefix
                                      + "> in qName " + qnameString
                                      + " is not mapped to a URI");
            } else {
                uri = pnp.getNamespace();
                assert uri == null || uri.length() > 0;
                if (lxprefix.length() == 0) {
                    if (uri == null) {
                        // (case 3)
                        prefix = null;
                        qName = localname;
                    } else {
                        // (case 4)
                        prefix = pnp.getPrefix();
                        qName = prefix + ":" + localname;
                    }
                } else {
                    if (uri == null) {
                        // (case 5)
                        prefix = null;
                        qName = localname;
                    } else {
                        // (case 6)
                        prefix = lxprefix;
                        // use equals() -- not interned yet
                        assert prefix.equals(pnp.getPrefix());
                        qName = qnameString;
                    }
                }
            }
            
            // Make sure everything is interned or null.  Some of
            // these tests are redundant, given that the assertion
            // below is true, but we haven't tested that yet, and
            // we're paranoid.
            //localname = (localname == null ? null : localname.intern());
            qName     = (qName == null     ? null : qName.intern());
            prefix    = (prefix == null    ? null : prefix.intern());
            uri       = (uri == null       ? null : uri.intern());
        }

        // localname and qName are always non-null
        // If not in a namespace, then prefix and uri are null
        // If in a namespace,     then prefix and uri are non-null,
        //                        and prefix is a prefix of qName
        // If they are not null,  then prefix and uri are not the empty string
        assert localname != null
                && qName != null
                && (prefix == null ? uri == null
                    : (uri != null
                       && prefix.length() > 0
                       && uri.length() > 0
                       && qName.startsWith(prefix)));
    }

    /**
     * Creates a new QName, by parsing the GI string passed to it.
     * Equivalent to {@link #QName(String,Stack,QName)}.
     *
     * @param qname a QName which is to be parsed
     * @param stack an element/namespace stack which provides context
     * for features and namespace mappings; this may not be null
     * @throws LxException if there is an error parsing the string,
     * if the stack was null, or if the QName was syntactically
     * invalid
     * @see #QName(String,Stack,QName)
     */
    public QName(String qname, Stack stack)
	    throws LxException {
        this(qname, stack, null);
    }

    /**
     * Make a QName consisting of a given NSAttName.  
     * This is for the use of the {@link #newNSAttribute} method.
     * @param xmlnsname a well-formed NSAttName
     * @throws LxException if the localname is malformed
     * @see #nsAttNamePattern
     */
    private QName(String xmlnsname) 
            throws LxException {
        if (nsAttNamePattern.matcher(xmlnsname).matches()) {
            prefix = null;
            uri = null;
            localname = xmlnsname.intern();
            qName = localname;
        } else {
            throw new LxException("Malformed NSAttName: " + xmlnsname);
        }
    }

    public static QName newNSAttribute(String prefix) 
            throws LxException {
        return new QName(prefix == null ? "xmlns" : "xmlns:"+prefix);
    }

    /**
     * Obtains the local name of this Element's GI.
     *
     * @return the element's GI as an XML LocalName, which is never null
     */
    public String localName() {
	return localname;
    }

    /**
     * Obtains the QName of this Element's GI.
     *
     * @return the element's GI as an XML QName, which is never null
     */
    public String qName() {
	return qName;
    }

    /**
     * Obtains the namespace of this QName
     *
     * @return the QName's namespace as a non-empty string, or null if it is
     * in no namespace
     */
    public String namespace() {
        return uri;
    }

    /**
     * Obtains the prefix of this QName
     * @return the QName's prefix as a non-empty string, or null if the QName
     * is in no namespace
     */
    public String prefix() {
        return prefix;
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("{");
        String ns = namespace();
        if (ns != null)
            sb.append(ns);
        sb.append("}").append(qName());
        return sb.toString();
    }

    /* Override Object methods */
    public boolean equals(Object o) {
        if (o == this)
            return true;
        if (! (o instanceof QName))
            return false;
        // rely on QName having interned instance variables
        QName qo = (QName)o;
        return (qo.localname == localname
                && qo.uri == uri);
    }

    public int hashCode() {
        // following is probably poor, however having the 2 here means that 
        // {}gi as a QName has a different hashCode from "gi" as a String
        // (there'll probably be instances of both in the map)
        return (uri == null ? 2 : uri.hashCode())
                * localname.hashCode();
    }

    /* clone does not need to be overridden */
}
