// This file is part of Lx.
//
// Copyright 2003, 2004, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.helpers;

import uk.me.nxg.lx.reader.LxReader;
import uk.me.nxg.lx.tokenizer.Tokenizer;
import uk.me.nxg.lx.tokenizer.Tokenizer.Token;

import java.util.List;
import java.util.ArrayList;

import junit.framework.Assert;
import junit.framework.TestCase;

public class LxWriterTest
        extends TestCase {

    java.util.regex.Pattern whitespace
            = java.util.regex.Pattern.compile("\\s+");

    public LxWriterTest(String name) {
        super(name);
    }

    public void testWriting()
            throws Exception {
        String[] testlx = {
            "(hello (gi1 a1: v1) (gi2 (gi3 a2: v2 a3: \"v3 v3a\" content)))",
            "(lx:namespace ((\"a\" \"urn:1\")) (gi1 a:a1: v1 x) (a:gi2 y))",
        };

        for (int i=0; i<testlx.length; i++) {
            LxReader lr = new LxReader();
            java.io.StringWriter sw = new java.io.StringWriter();
            LxWriter lw = new LxWriter(sw);

            lr.setContentHandler(lw);
            lr.parse(new org.xml.sax.InputSource
                     (new java.io.StringReader(testlx[i])));

            String res = sw.toString();
            //System.err.println("res=" + res);
            assertEquals(whitespace.matcher(testlx[i]).replaceAll(""),
                         whitespace.matcher(res).replaceAll(""));
        }
    }

    public void testCharacterEscaping()
            throws Exception {
        String[] testpairs = {
            // not escaped
            "plain", "plain",
            "a-b%!=x$+*", "a-b%!=x$+*",
            "x'", "x'",

            // escaped with "..."
            "two words", "\"two words\"",
            "(hello)", "\"(hello)\"",
            "{", "\"{\"",
            "|", "\"|\"",
            "#f", "\"#f\"",
            "hello:", "\"hello:\"",
            "esc\\ape", "\"esc\\\\ape\"",
            "x[b]", "\"x[b]\"",
            "x\\[b]", "\"x\\\\[b]\"",
            "x]", "\"x]\"",
            "x[]]", "\"x[]]\"",
            "x[[]", "\"x[[]\"",
            "x\"", "\"x\\\"\"",

            // with newlines
            "plain\n", "\"plain\n\"",
            "two words\n", "\"two words\n\"",
            "esc\\ape\n", "\"esc\\\\ape\n\"",
            "x[b]\n", "\"x[b]\n\"",
            "x\\[b]\n", "\"x\\\\[b]\n\"",
            "x]\n", "\"x]\n\"",
            "x[]]\n", "\"x[]]\n\"",
            "x[[]\n", "\"x[[]\n\"",
            "x\"\n", "\"x\\\"\n\"",
            "x'\n", "\"x'\n\"",

            /*
              The following is how these would be escaped with [...]
            // with newlines -- escape with [...]
            "plain\n", "[plain\n]",
            "two words\n", "[two words\n]",
            "esc\\ape\n", "[esc\\\\ape\n]",
            "x[b]\n", "[x[b]\n]",
            "x\\[b]\n", "[x\\\\[b]\n]",
            "x]\n", "[x\\]\n]",
            "x[]]\n", "[x\\[\\]\\]\n]",
            "x[[]\n", "[x\\[\\[\\]\n]",
            "x\"\n", "[x\"\n]",
            "x'\n", "[x'\n]",
            */
        };

        java.io.StringWriter sw = new java.io.StringWriter();
        StringBuffer target = new StringBuffer();
        LxWriter lw = new LxWriter(sw);
        lw.setTrimWhitespace(false);
        lw.startDocument();
        for (int i=0; i<testpairs.length; i+=2) {
            lw.characters
                    (testpairs[i].toCharArray(), 0, testpairs[i].length());
            target.append(' ').append(testpairs[i+1]);
        }
        lw.endDocument();

        assertEquals(target.toString(), sw.toString());
    }
    
    // Following test is redundant with previous (since both
    // LxWriter.characters and LxWriter.startElement use the same
    // underlying method to escape strings), and is also more
    // intricate.  We could afford to lose it.
    public void testAttributeEscaping()
            throws Exception {
        List tests = new ArrayList();
        List results = new ArrayList();

        // not quoted
        tests.add("plain");
        results.add("plain");

        tests.add("x'y'");
        results.add("x'y'");

        // quoted
        tests.add("two words");
        results.add("\"two words\"");

        tests.add("esc\\ape");
        results.add("\"esc\\\\ape\"");

        tests.add("x[b]");
        results.add("\"x[b]\"");

        tests.add("x\\[b]");
        results.add("\"x\\\\[b]\"");

        tests.add("x]");
        results.add("\"x]\"");

        org.xml.sax.helpers.AttributesImpl ai
                = new org.xml.sax.helpers.AttributesImpl();
        // sb holds what the output Lx script should look like
        StringBuffer sb = new StringBuffer("(gi ");
        for (int i=0; i<tests.size(); i++) {
            String attname = "a" + Integer.toString(i);
            ai.addAttribute("", "", attname, "CDATA", (String)tests.get(i));
            sb.append(' ')
                    .append(attname)
                    .append(": ")
                    .append((String)results.get(i));
        }
        sb.append(')');

        java.io.StringWriter sw = new java.io.StringWriter();
        LxWriter lw = new LxWriter(sw);
        lw.startElement("", "", "gi", ai);
        lw.endElement("", "", "gi");

        String testResult = sb.toString(); // what it should look like
        String writeResult = sw.toString(); // output from LxWriter
        
        assertEquals(whitespace.matcher(testResult).replaceAll(" "),
                     whitespace.matcher(writeResult).replaceAll(" "));

        // now read in the testResult, and check that the sequence of
        // tokens matches the original strings in the list tests[]
        Tokenizer t = new Tokenizer(testResult);
        assertEquals(Token.OPENPAREN, t.getToken());
        assertToken(Token.SYMBOL, "gi", t);
        for (int i=0; i<tests.size(); i++) {
            assertEquals(Token.KEYWORD, t.getToken());
            t.getToken();
            assertEquals((String)tests.get(i), t.value());
        }
        
    }
    
    private void assertToken(Token expectedToken, String expectedValue,
			     Tokenizer t) 
	    throws Exception {
	assertEquals(expectedToken, t.getToken());
	assertEquals(expectedValue, t.value());
    }
}

