// This file is part of Lx.
//
// Copyright 2003, 2004, 2005, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.reader;

import uk.me.nxg.lx.TraceContentHandler;

import junit.framework.Assert;
import junit.framework.TestCase;

import java.io.StringReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.util.List;
import java.util.LinkedList;

import org.xml.sax.InputSource;
import org.xml.sax.SAXParseException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;

public class LxReaderTest
	extends TestCase {

    public LxReaderTest(String name) {
	super(name);
    }

    public void testEvents()
	    throws Exception {

	String testlx = "(hello there (how \"goes it\" with (you)) today)";
	LxReader h = new LxReader();
	List eventlist = new LinkedList();
	
	h.setContentHandler(new TraceContentHandler(eventlist));
	h.parse(new InputSource(new StringReader(testlx)));

        String[] expectedStrings = {
            "startDocument",
            "startElement {}hello/hello[]",
            "characters 5:there",
            "startElement {}how/how[]",
            "characters 7:goes it",
            "characters 4:with",
            "startElement {}you/you[]",
            "endElement {}you/you",
            "endElement {}how/how",
            "characters 5:today",
            "endElement {}hello/hello",
            "endDocument",
        };
        assertListEquals(expectedStrings, eventlist);
    }

    public void testUnicode()
            throws Exception {
        String testlx = "(hello \"my little \u00e9clair\u2013gem\u00fcse\")";
        LxReader h = new LxReader();
	List eventlist = new LinkedList();
	
	h.setContentHandler(new TraceContentHandler(eventlist));
	h.parse(new InputSource(new StringReader(testlx)));

        String[] expectedStrings = {
            "startDocument",
            "startElement {}hello/hello[]",
            "characters 23:my little \u00e9clair\u2013gem\u00fcse",
            "endElement {}hello/hello",
            "endDocument",
        };
        assertListEquals(expectedStrings, eventlist);
    }

    public void testElements()
	    throws Exception {
	String[] testlx = {
            "(hello (gi1 a1: v1) (gi2 (gi3 a2: v2 a3: v3 content)))",
            "(hello (gi1 (@ (a1 v1))) (gi2 (gi3 (@ (a2 v2) (a3 v3)) content)))"
        };
        String[] expectedStrings = {
            "startDocument",
            "startElement {}hello/hello[]",
            "startElement {}gi1/gi1[a1=v1]",
            "endElement {}gi1/gi1",
            "startElement {}gi2/gi2[]",
            "startElement {}gi3/gi3[a2=v2,a3=v3]",
            "characters 7:content",
            "endElement {}gi3/gi3",
            "endElement {}gi2/gi2",
            "endElement {}hello/hello",
            "endDocument",
        };

        for (int i=0; i<testlx.length; i++) {
            LxReader h = new LxReader();
            List eventlist = new LinkedList();

            assertFalse(h.getFeature(LxReader.SAX_NAMESPACE_PREFIXES));

            h.setContentHandler(new TraceContentHandler(eventlist));
            h.parse(new InputSource(new StringReader(testlx[i])));

            assertListEquals(expectedStrings, eventlist);
        }
    }

    public void testElementsNS()
            throws Exception {
        String testlx 
                = "(lx:namespace ((\"a\" \"urn:1\")"
                + "               (\"b\" \"urn:2\"))"
                + "  (a:gi1 a1: v1 (a:gi2 a2: v2 a:a3: v3 text)"
                + "                (b:gi2 a2: v2)"
                + "                (gi2 a2: v2))"
                + ")";
        LxReader h = new LxReader();
        List eventlist = new LinkedList();

        // Case A: default behaviour
        // Not quite -- don't make any assertions about the default,
        // since I'm experimenting with changing this
        assertFalse(h.getFeature(LxReader.SAX_NAMESPACE_PREFIXES));
        h.setFeature(LxReader.LX_ATTRIBUTES_IN_NO_NAMESPACE, false);
        h.setFeature(LxReader.LX_ATTRIBUTES_INHERIT_NAMESPACE, false);
        //assertFalse(h.getFeature(LxReader.LX_ATTRIBUTES_INHERIT_NAMESPACE));

        TraceContentHandler tch = new TraceContentHandler(eventlist);
        h.setContentHandler(tch);
        h.parse(new InputSource(new StringReader(testlx)));

        String[] expectedStringsA = {
            "startDocument",
            "map+ a>urn:1",
            "map+ b>urn:2",
            "startElement {urn:1}gi1/a:gi1[a1=v1]",
            "startElement {urn:1}gi2/a:gi2[a2=v2,{urn:1}a3/a:a3=v3]",
            "characters 4:text",
            "endElement {urn:1}gi2/a:gi2",
            "startElement {urn:2}gi2/b:gi2[a2=v2]",
            "endElement {urn:2}gi2/b:gi2",
            "startElement {}gi2/gi2[a2=v2]",
            "endElement {}gi2/gi2",
            "endElement {urn:1}gi1/a:gi1",
            "map- b",
            "map- a",
            "endDocument",
        };
        assertListEquals(expectedStringsA, eventlist);

        // Case B: Test handling of attribute qnames when the inherit-namespace
        // feature is true 
        h.setFeature(LxReader.LX_ATTRIBUTES_IN_NO_NAMESPACE, false);
        h.setFeature(LxReader.LX_ATTRIBUTES_INHERIT_NAMESPACE, true);
        h.parse(new InputSource(new StringReader(testlx)));
        String[] expectedStringsB = {
            "startDocument",
            "map+ a>urn:1",
            "map+ b>urn:2",
            "startElement {urn:1}gi1/a:gi1[{urn:1}a1/a:a1=v1]",
            "startElement {urn:1}gi2/a:gi2[{urn:1}a2/a:a2=v2,{urn:1}a3/a:a3=v3]",
            "characters 4:text",
            "endElement {urn:1}gi2/a:gi2",
            "startElement {urn:2}gi2/b:gi2[{urn:2}a2/b:a2=v2]",
            "endElement {urn:2}gi2/b:gi2",
            "startElement {}gi2/gi2[a2=v2]",
            "endElement {}gi2/gi2",
            "endElement {urn:1}gi1/a:gi1",
            "map- b",
            "map- a",
            "endDocument",
        };
        assertListEquals(expectedStringsB, eventlist);

        // More subtle ones involving null prefixes and namespaces
        testlx 
                = "(lx:namespace ((\"a\" #f)"
                + "               (#f \"urn:2\"))"
                + "  (a:gi1 a:a1: v1 a2: v2)"
                + "  (gi1   a:a1: v1 a2: v2)"
                + ")";
        // Case C: default behaviour again.
        // First set inherit-namespace to (default) false
        h.setFeature(LxReader.LX_ATTRIBUTES_INHERIT_NAMESPACE, false);
        h.parse(new InputSource(new StringReader(testlx)));
        String[] expectedStringsC = {
            "startDocument",
            "map+ lxns>urn:2",
            "startElement {}gi1/gi1[a1=v1,{urn:2}a2/lxns:a2=v2]",
            "endElement {}gi1/gi1",
            "startElement {urn:2}gi1/lxns:gi1[a1=v1,{urn:2}a2/lxns:a2=v2]",
            "endElement {urn:2}gi1/lxns:gi1",
            "map- lxns",
            "endDocument",
        };
        assertListEquals(expectedStringsC, eventlist);

        // Case D: with inherit-namespace true
        h.setFeature(LxReader.LX_ATTRIBUTES_INHERIT_NAMESPACE, true);
        h.parse(new InputSource(new StringReader(testlx)));
        String[] expectedStringsD = {
            "startDocument",
            "map+ lxns>urn:2",
            "startElement {}gi1/gi1[a1=v1,a2=v2]",
            "endElement {}gi1/gi1",
            "startElement {urn:2}gi1/lxns:gi1[a1=v1,{urn:2}a2/lxns:a2=v2]",
            "endElement {urn:2}gi1/lxns:gi1",
            "map- lxns",
            "endDocument",
        };
        assertListEquals(expectedStringsD, eventlist);

        // Case E: inherit-namespace false, in-no-namespace true
        h.setFeature(LxReader.LX_ATTRIBUTES_IN_NO_NAMESPACE, true);
        h.setFeature(LxReader.LX_ATTRIBUTES_INHERIT_NAMESPACE, false);
        h.parse(new InputSource(new StringReader(testlx)));
        String[] expectedStringsE = {
            "startDocument",
            "map+ lxns>urn:2",
            "startElement {}gi1/gi1[a1=v1,a2=v2]",
            "endElement {}gi1/gi1",
            "startElement {urn:2}gi1/lxns:gi1[a1=v1,a2=v2]",
            "endElement {urn:2}gi1/lxns:gi1",
            "map- lxns",
            "endDocument",
        };
        assertListEquals(expectedStringsE, eventlist);

        // Case F: The same as before, but with the lx:feature construction.
        // First reset the parser to the default set of features.
        h.setFeature(LxReader.LX_ATTRIBUTES_INHERIT_NAMESPACE, false);
        h.setFeature(LxReader.LX_ATTRIBUTES_IN_NO_NAMESPACE, false);
        testlx 
                = "(lx:namespace ((\"a\" #f)"
                + "               (#f \"urn:2\"))"
                + "  (a:gi1 a:a1: v1 a2: v2)"
                + "  (gi1   a:a1: v1 a2: v2)"

                + "  (lx:feature ((\"http://ns.nxg.me.uk/lx/attributes-inherit-namespace\" #t))"
                + "     (a:gi1 a:a1: v1 a2: v2)"
                + "     (gi1   a:a1: v1 a2: v2)"

                // can set features true just by naming them
                + "     (lx:feature (\"http://ns.nxg.me.uk/lx/attributes-in-no-namespace\""
                + "                  (\"http://ns.nxg.me.uk/lx/attributes-inherit-namespace\" #f))"
                + "        (a:gi1 a:a1: v1 a2: v2)"
                + "        (gi1   a:a1: v1 a2: v2)"

                + ")))";
        h.parse(new InputSource(new StringReader(testlx)));
        String[] expectedStringsF = {
            "startDocument",
            "map+ lxns>urn:2",

            // first as in case C
            "startElement {}gi1/gi1[a1=v1,{urn:2}a2/lxns:a2=v2]",
            "endElement {}gi1/gi1",
            "startElement {urn:2}gi1/lxns:gi1[a1=v1,{urn:2}a2/lxns:a2=v2]",
            "endElement {urn:2}gi1/lxns:gi1",

            // then as in case D
            "startElement {}gi1/gi1[a1=v1,a2=v2]",
            "endElement {}gi1/gi1",
            "startElement {urn:2}gi1/lxns:gi1[a1=v1,{urn:2}a2/lxns:a2=v2]",
            "endElement {urn:2}gi1/lxns:gi1",

            // finally as in case E
            "startElement {}gi1/gi1[a1=v1,a2=v2]",
            "endElement {}gi1/gi1",
            "startElement {urn:2}gi1/lxns:gi1[a1=v1,a2=v2]",
            "endElement {urn:2}gi1/lxns:gi1",

            "map- lxns",
            "endDocument",
        };
        assertListEquals(expectedStringsF, eventlist);
    }

    public void testElementsMultinamespace()
            throws Exception {
        String testlx
                = "(lx:namespace ((\"a\" \"http://x.org/a\")"
                + "               (\"b\" \"http://x.org/b\")"
                + "               (\"html\" #f))"
                + "  (gi1 a1: v1 (a:gi2 b:a2: v2 text)"
                + "              (html:h1)"
                + "       (lx:namespace ((\"c\" \"http://x.org/c\")"
                + "                      (\"b\" \"http://x.org/d\"))"
                + "          (c:gi3 b:a3: v3))))";
        LxReader h = new LxReader();
        List eventlist = new LinkedList();

        assertFalse(h.getFeature(LxReader.SAX_NAMESPACE_PREFIXES));

        TraceContentHandler tch = new TraceContentHandler(eventlist);
        h.setContentHandler(tch);
        h.parse(new InputSource(new StringReader(testlx)));

        String[] expectedStrings = {
            "startDocument",
            "map+ a>http://x.org/a",
            "map+ b>http://x.org/b",
            "startElement {}gi1/gi1[a1=v1]",
            "startElement {http://x.org/a}gi2/a:gi2[{http://x.org/b}a2/b:a2=v2]",
            "characters 4:text",
            "endElement {http://x.org/a}gi2/a:gi2",
            "startElement {}h1/h1[]",
            "endElement {}h1/h1",
            "map+ c>http://x.org/c",
            "map+ b>http://x.org/d",
            "startElement {http://x.org/c}gi3/c:gi3[{http://x.org/d}a3/b:a3=v3]",
            "endElement {http://x.org/c}gi3/c:gi3",
            "map- b",
            "map- c",
            "endElement {}gi1/gi1",
            "map- b",
            "map- a",
            "endDocument",
        };
        assertListEquals(expectedStrings, eventlist);


        // Now same again, but this time with namespace-prefixes true
        LxReader h2 = new LxReader();
        h2.setFeature(LxReader.SAX_NAMESPACE_PREFIXES, true);
        tch = new TraceContentHandler(eventlist);
        h2.setContentHandler(tch);
        h2.parse(new InputSource(new StringReader(testlx)));

        String[] nsexpectedStrings = {
            "startDocument",
            "map+ a>http://x.org/a",
            "map+ b>http://x.org/b",
            "startElement {}gi1/gi1[NS:xmlns:a=http://x.org/a,NS:xmlns:b=http://x.org/b,a1=v1]",
            "startElement {http://x.org/a}gi2/a:gi2[{http://x.org/b}a2/b:a2=v2]",
            "characters 4:text",
            "endElement {http://x.org/a}gi2/a:gi2",
            "startElement {}h1/h1[]",
            "endElement {}h1/h1",
            "map+ c>http://x.org/c",
            "map+ b>http://x.org/d",
            "startElement {http://x.org/c}gi3/c:gi3[NS:xmlns:b=http://x.org/d,NS:xmlns:c=http://x.org/c,{http://x.org/d}a3/b:a3=v3]",
            "endElement {http://x.org/c}gi3/c:gi3",
            "map- b",
            "map- c",
            "endElement {}gi1/gi1",
            "map- b",
            "map- a",
            "endDocument",
        };
        assertListEquals(nsexpectedStrings, eventlist);
    }
    
    public void testNSMangling()
            throws Exception {
        String testlx 
                = "(lx:namespace ((a \"x:1\")"
                + "               (#f \"x:2\")"
                + "               (\"\" \"x:3\")"
                + "               (b \"\")"
                + "               (c #f)"
                + "               (#f #f))"
                + " (a:e1 (e2) (:e3))"
                + " (b:e4 (c:e5)))"
                + "(lx:namespace ((#f #f))" // no-op: tests that this is indeed
                			    // skipped, and that a null set of 
                			    // namespace mappings is OK
                + " (e6))";

        LxReader h = new LxReader();
        List eventlist = new LinkedList();

        assertFalse(h.getFeature(LxReader.SAX_NAMESPACE_PREFIXES));

        // force the prefix-generation prefix
        h.setProperty(LxReader.LX_GENERATED_PREFIX, "gen");

        TraceContentHandler tch = new TraceContentHandler(eventlist);
        tch.showQNames(false);
        h.setContentHandler(tch);
        h.parse(new InputSource(new StringReader(testlx)));

        String[] expectedStrings = {
            "startDocument",
            "map+ a>x:1",
            "map+ gen>x:2",
            "map+ gen1>x:3",
            "startElement {x:1}e1[]",
            "startElement {x:2}e2[]",
            "endElement {x:2}e2",
            "startElement {x:3}e3[]",
            "endElement {x:3}e3",
            "endElement {x:1}e1",
            "startElement {}e4[]",
            "startElement {}e5[]",
            "endElement {}e5",
            "endElement {}e4",
            "map- gen1",
            "map- gen",
            "map- a",
            "startElement {}e6[]",
            "endElement {}e6",
            "endDocument",
        };
        assertListEquals(expectedStrings, eventlist);

        // Now same again, but this time with namespace-prefixes true
        h = new LxReader();
        h.setFeature(LxReader.SAX_NAMESPACE_PREFIXES, true);
        // force the prefix-generation prefix
        h.setProperty(LxReader.LX_GENERATED_PREFIX, "gen");
        tch = new TraceContentHandler(eventlist);
        tch.showQNames(false);
        h.setContentHandler(tch);
        h.parse(new InputSource(new StringReader(testlx)));

        String[] nsExpectedStrings = {
            "startDocument",
            "map+ a>x:1",
            "map+ gen>x:2",
            "map+ gen1>x:3",
            "startElement {x:1}e1[NS:xmlns:a=x:1,NS:xmlns:gen1=x:3,NS:xmlns:gen=x:2]",
            "startElement {x:2}e2[]",
            "endElement {x:2}e2",
            "startElement {x:3}e3[]",
            "endElement {x:3}e3",
            "endElement {x:1}e1",
            "startElement {}e4[]",
            "startElement {}e5[]",
            "endElement {}e5",
            "endElement {}e4",
            "map- gen1",
            "map- gen",
            "map- a",
            "startElement {}e6[]",
            "endElement {}e6",
            "endDocument",
        };
        assertListEquals(nsExpectedStrings, eventlist);
    }

    public void testPropertySetting()
            throws Exception {
        String testlx 
                = "(lx:property ((\"http://ns.nxg.me.uk/lx/generated-prefix\""
                + "               \"gen\"))"
                + "  (lx:namespace ((\"a\" #f)"
                + "                 (#f \"urn:2\"))"
                + "    (a:gi1 a:a1: v1 a2: v2)"
                + "    (gi1   a:a1: v1 a2: v2)"
                + "))";

        LxReader h = new LxReader();
        List eventlist = new LinkedList();

        TraceContentHandler tch = new TraceContentHandler(eventlist);
        h.setContentHandler(tch);
        h.parse(new InputSource(new StringReader(testlx)));

        String[] expectedStrings = {
            "startDocument",
            "map+ gen>urn:2",
            "startElement {}gi1/gi1[a1=v1,a2=v2]",
            "endElement {}gi1/gi1",
            "startElement {urn:2}gi1/gen:gi1[a1=v1,a2=v2]",
            "endElement {urn:2}gi1/gen:gi1",
            "map- gen",
            "endDocument",
        };
        assertListEquals(expectedStrings, eventlist);
    }

    public void testDTD() 
            throws Exception {

        LxReader h = new LxReader();
        List eventlist = new LinkedList();
        TraceContentHandler tch = new TraceContentHandler(eventlist);
        tch.showQNames(false);
        h.setContentHandler(tch);

        String testlx
                = "(lx:dtd ((gi1 (att11 #t)"
                + "              (att12 #f))"
                + "         (gi2 (att21 #f)))"
                + "   (gi1 att11: v11 text)" // req'd att named
                + "   (gi1 v11 text)" // req'd att not named

                + "   (lx:dtd ((gi1 #f))" // inner dtd -- turns off checking
                + "      (gi1 attxx: x))"

                // following four should be identical
                + "   (gi1 att12: v12 v11)" // optional then implicit req'd
                + "   (gi1 v11 att12: v12)" // implicit req'd then optional
                + "   (gi1 att12: v12 att11: v11)" // both named
                + "   (gi1 (@ (att12 v12) (att11 v11)))"

                + "   (gi2 att21: v21))" // simple
                ;
        h.parse(new InputSource(new StringReader(testlx)));
        String[] expectedStringsA = {
            "startDocument",

            // next two should be identical
            "startElement {}gi1[att11=v11]",
            "characters 4:text",
            "endElement {}gi1",
            "startElement {}gi1[att11=v11]",
            "characters 4:text",
            "endElement {}gi1",

            // inner dtd
            "startElement {}gi1[attxx=x]",
            "endElement {}gi1",

            // following four should be identical
            "startElement {}gi1[att11=v11,att12=v12]",
            "endElement {}gi1",
            "startElement {}gi1[att11=v11,att12=v12]",
            "endElement {}gi1",
            "startElement {}gi1[att11=v11,att12=v12]",
            "endElement {}gi1",
            "startElement {}gi1[att11=v11,att12=v12]",
            "endElement {}gi1",

            // simple
            "startElement {}gi2[att21=v21]",
            "endElement {}gi2",

            "endDocument",
        };
        assertListEquals(expectedStringsA, eventlist);

        testlx  = "(lx:dtd ((gi (a1 #t)"
                + "             (a2 #t)"
                + "             (a3 #f)"
                + "             (a4 #f)))"
                // Interleave implicit and explicit attributes
                + "  (gi v1 a3: v3 v2 a4: v4))";
        h.parse(new InputSource(new StringReader(testlx)));
        String[] expectedStringsB = {
            "startDocument",
            "startElement {}gi[a1=v1,a2=v2,a3=v3,a4=v4]",
            "endElement {}gi",
            "endDocument",
        };
        assertListEquals(expectedStringsB, eventlist);
    }

    public void testDTDWithNamespaces() 
            throws Exception {
        String testlx
                = "(lx:feature ((\""
                +                 LxReader.LX_ATTRIBUTES_IN_NO_NAMESPACE
                +                "\" #f))"
                + "(lx:namespace ((a \"urn:1\")"
                + "               (#f \"urn:2\"))"
                // all required attributes
                + " (lx:dtd ((a:gi1 (a:a1 #t) (a2 #t))"
                + "          (gi2 (b1 #t)))"
                + "  (lx:namespace ((b \"urn:1\"))"
                + "   (b:gi1 b:a1: va1 a2: va2)"
                + "   (b:gi1 va1 va2))"
                + "  (gi2 b1: vb1)"
                + "  (gi2 vb1))))"

                // same again, but with the not-in-namespaces feature true
                + "(lx:feature (\""
                +                 LxReader.LX_ATTRIBUTES_IN_NO_NAMESPACE
                +                "\")"
                + " (lx:namespace ((a \"urn:1\")"
                + "                (#f \"urn:2\"))"
                + "  (lx:dtd ((a:gi1 (a:a1 #t) (a2 #t))"
                + "           (gi2 (b1 #t)))"
                + "   (a:gi1 a:a1: va1 a2: va2)"
                + "   (a:gi1 va1 va2)"
                + "   (gi2 b1: vb1)"
                + "   (gi2 vb1))))";

        LxReader h = new LxReader();
        List eventlist = new LinkedList();

        TraceContentHandler tch = new TraceContentHandler(eventlist);
        tch.showQNames(false);
        h.setContentHandler(tch);
        h.parse(new InputSource(new StringReader(testlx)));

        String[] expectedStringsA = {
            "startDocument",

            "map+ a>urn:1",
            "map+ lxns>urn:2",
            "map+ b>urn:1",

            "startElement {urn:1}gi1[{urn:1}a1=va1,{urn:2}a2=va2]",
            "endElement {urn:1}gi1",
            "startElement {urn:1}gi1[{urn:1}a1=va1,{urn:2}a2=va2]",
            "endElement {urn:1}gi1",

            "map- b",

            "startElement {urn:2}gi2[{urn:2}b1=vb1]",
            "endElement {urn:2}gi2",
            "startElement {urn:2}gi2[{urn:2}b1=vb1]",
            "endElement {urn:2}gi2",

            "map- lxns",
            "map- a",

            "map+ a>urn:1",
            "map+ lxns>urn:2",

            "startElement {urn:1}gi1[a2=va2,{urn:1}a1=va1]",
            "endElement {urn:1}gi1",
            "startElement {urn:1}gi1[a2=va2,{urn:1}a1=va1]",
            "endElement {urn:1}gi1",

            "startElement {urn:2}gi2[b1=vb1]",
            "endElement {urn:2}gi2",
            "startElement {urn:2}gi2[b1=vb1]",
            "endElement {urn:2}gi2",

            "map- lxns",
            "map- a",

            "endDocument",
        };
        assertListEquals(expectedStringsA, eventlist);
    }

    public void testDTDFailures() {
	String[] badInputs = {
            "(lx:dtd ((gi1)) (gi1 att: val))", // no valid atts
            "(lx:dtd ((gi1 att1)) (gi1 att2: val2))", // invalid att
            "(lx:dtd ((gi1 (att1 #t))) (gi1))", // missing required att

            // as above, but with (@...) syntax
            "(lx:dtd ((gi1)) (gi1 (@ (att val))))",
            "(lx:dtd ((gi1 att1)) (gi1 (@ (att2 val2))))",
        };

        for (int i=0; i<badInputs.length; i++) {
	    try {
		LxReader r = new LxReader();
		r.setContentHandler(new org.xml.sax.helpers.DefaultHandler());
		r.parse(new InputSource(new StringReader(badInputs[i])));
		fail("LxReader didn't fail \"" + badInputs[i] + "\"");
	    } catch (SAXParseException e) {
		// OK
	    } catch (Exception e) {
		fail("LxReader produced unexpected exception for input \""
		     + badInputs[i] + "\": " + e);
	    }
	}
    }

    public void testLet()
            throws Exception {
        // test the lx:let construct
        String testlx
                = "(lx:let ((namespace (\"x\" \"urn:1\"))"
                + "         (dtd (\"gi1\" \"a1\" (\"a2\" #t))"
                + "              (\"x:gix\" \"x:ax\")))"
                + "  (gi1 v2)"
                + "  (x:gix x:ax: vx))";
        LxReader h = new LxReader();
        List eventlist = new LinkedList();

        TraceContentHandler tch = new TraceContentHandler(eventlist);
        tch.showQNames(false);
        h.setContentHandler(tch);
        h.parse(new InputSource(new StringReader(testlx)));

        String[] expectedStrings = {
            "startDocument",
            "map+ x>urn:1",
            "startElement {}gi1[a2=v2]",
            "endElement {}gi1",
            "startElement {urn:1}gix[{urn:1}ax=vx]",
            "endElement {urn:1}gix",
            "map- x",
            "endDocument",
        };
        assertListEquals(expectedStrings, eventlist);
    }

    public void testAtAttributeEdges()
            throws Exception {
        String testlx = "(gi1 (@))";
        LxReader h = new LxReader();
        List eventlist = new LinkedList();

        TraceContentHandler tch = new TraceContentHandler(eventlist);
        h.setContentHandler(tch);
        h.parse(new InputSource(new StringReader(testlx)));

        String[] expectedStrings = {
            "startDocument",
            "startElement {}gi1/gi1[]",
            "endElement {}gi1/gi1",
            "endDocument",
        };
        assertListEquals(expectedStrings, eventlist);
    }

    public void testParseFailures() {
	String[] badInputs
		= { "(el (el)", // too few right parentheses
		    "(el (el)))",	// too many
		    "(el (unclosed \"string'))", // unclosed quote
		    "(el ((inner list)))", // directly enclosed list
		    "(el single-keyword:)", // no keyword value
		    "(el keyword: #badvalue)", // improper value
                    "(el keyword: #\"thing\")", // wrong type of token value
                    "(el keyword: #(oops))",    // wrong type of token value
                    "(el content keyword:)", // keyword out of context
                    "(el content #t)", // token out of context
                    "(el att: value att: value)", // att specified twice
                    "(el att: (list value))",     // att requires string or symbol value
                    "(el att: value (@ (att2 value2)))", // both att forms
                    "(el (@ (att2 value2)) att: value)", // both att forms
                    "(el (@ (a1 v1)) (@ (a2 v2)))", // duplicate atts
                    "(el content (@ (a1 v1)))", // atts in content
                    "(el (@ nothing))", // non-list in atts
                    "(el (@ (hello there) nothing))", // non-list in atts
                    "(el (@ (hello)))", // non-pair in atts
                    "(el (@ (hello there now)))", // too many items in atts
                    "(el (@ (att val) (att val)))", // att specified twice
                    "(el1 (@@ (hello there)))", // not attlist, not QName
                    "(\"el\" content)",         // wrong type for element name
                    "((el) content)",           // ditto
	          };

	for (int i=0; i<badInputs.length; i++) {
	    try {
		LxReader r = new LxReader();
		r.setContentHandler(new org.xml.sax.helpers.DefaultHandler());
		r.parse(new InputSource(new StringReader(badInputs[i])));
		fail("LxReader didn't fail \"" + badInputs[i] + "\"");
	    } catch (SAXParseException e) {
		// OK
	    } catch (Exception e) {
                e.printStackTrace(System.err);
		fail("LxReader produced unexpected exception for input \""
		     + badInputs[i] + "\": " + e);
	    }
	}
    }

    public void testFeatures()
            throws Exception {
        LxReader r = new LxReader();
        String ns = "http://xml.org/sax/features/namespaces";
        String nsp = "http://xml.org/sax/features/namespace-prefixes";

        assertTrue(r.getFeature(ns));
        assertFalse(r.getFeature(nsp));

        r.setFeature(ns, true);

        r.setFeature(ns, false);
        assertFalse(r.getFeature(ns));
        
        r.setFeature(nsp, false);

        class Pair {
            private String s;
            private boolean b;
            private Class c;
            Pair(String s, boolean b, Class c) {
                this.s = s;
                this.b = b;
                this.c = c;
            }
            void trySetting(LxReader r) {
                try {
                    r.setFeature(s, b);
                    fail("Setting feature " + s
                         + " to value " + b
                         + " did not fail");
                } catch (Exception e) {
                    if (c.isInstance(e)) {
                        // OK
                    } else {
                        fail("Setting feature " + s
                             + " to value " + b
                             + " produced unexpected exception " + e);
                    }
                }
            }
        }
        Pair[] featureset = {
            //new Pair(ns, false, SAXNotSupportedException.class),
            //new Pair(nsp, true, SAXNotSupportedException.class),
            new Pair("wibble", true, SAXNotRecognizedException.class),
        };
        for (int i=0; i<featureset.length; i++) {
            featureset[i].trySetting(r);
        }
    }


    /* ******************** Helper classes ******************** */

    private void assertListEquals(String[] expected, List l) {
        for (int i=0; i<expected.length; i++) {
            String head = (String)l.remove(0);
            assertNotNull(head);
            assertEquals(expected[i], head);
        }
        assertEquals(0, l.size());
    }

    /** 
     * Assert that a condition is false.
     * Older versions of JUnit don't have this. 
     */
    static public void assertFalse(boolean test) {
        assertTrue(! test);
    }
}
